@@ -178,77 +178,6 @@ _clutter_stage_window_clear_update_time (ClutterStageWindow *window)
   iface->clear_update_time (window);
 }
 
-void
-_clutter_stage_window_add_redraw_clip (ClutterStageWindow    *window,
-                                       cairo_rectangle_int_t *stage_clip)
-{
-  ClutterStageWindowInterface *iface;
-
-  g_return_if_fail (CLUTTER_IS_STAGE_WINDOW (window));
-
-  iface = CLUTTER_STAGE_WINDOW_GET_IFACE (window);
-  if (iface->add_redraw_clip != NULL)
-    iface->add_redraw_clip (window, stage_clip);
-}
-
-/* Determines if the backend will clip the rendering of the next
- * frame.
- *
- * Note: at the start of each new frame there is an implied clip that
- * clips everything (i.e. nothing would be drawn) so this function
- * will return True at the start of a new frame if the backend
- * supports clipped redraws.
- */
-gboolean
-_clutter_stage_window_has_redraw_clips (ClutterStageWindow *window)
-{
-  ClutterStageWindowInterface *iface;
-
-  g_return_val_if_fail (CLUTTER_IS_STAGE_WINDOW (window), FALSE);
-
-  iface = CLUTTER_STAGE_WINDOW_GET_IFACE (window);
-  if (iface->has_redraw_clips != NULL)
-    return iface->has_redraw_clips (window);
-
-  return FALSE;
-}
-
-/* Determines if the backend will discard any additional redraw clips
- * and instead promote them to a full stage redraw.
- *
- * The ideas is that backend may have some heuristics that cause it to
- * give up tracking redraw clips so this can be used to avoid the cost
- * of calculating a redraw clip when we know it's going to be ignored
- * anyway.
- */
-gboolean
-_clutter_stage_window_ignoring_redraw_clips (ClutterStageWindow *window)
-{
-  ClutterStageWindowInterface *iface;
-
-  g_return_val_if_fail (CLUTTER_IS_STAGE_WINDOW (window), FALSE);
-
-  iface = CLUTTER_STAGE_WINDOW_GET_IFACE (window);
-  if (iface->ignoring_redraw_clips != NULL)
-    return iface->ignoring_redraw_clips (window);
-
-  return TRUE;
-}
-
-cairo_region_t *
-_clutter_stage_window_get_redraw_clip (ClutterStageWindow *window)
-{
-  ClutterStageWindowInterface *iface;
-
-  g_return_val_if_fail (CLUTTER_IS_STAGE_WINDOW (window), FALSE);
-
-  iface = CLUTTER_STAGE_WINDOW_GET_IFACE (window);
-  if (iface->get_redraw_clip != NULL)
-    return iface->get_redraw_clip (window);
-
-  return NULL;
-}
-
 void
 _clutter_stage_window_set_accept_focus (ClutterStageWindow *window,
                                         gboolean            accept_focus)
