@@ -22,8 +22,9 @@
 #ifndef META_CURSOR_H
 #define META_CURSOR_H
 
-#include <meta/common.h>
-#include <meta/boxes.h>
+#include "backends/meta-backend-types.h"
+#include "meta/common.h"
+#include "meta/boxes.h"
 
 #define META_TYPE_CURSOR_SPRITE (meta_cursor_sprite_get_type ())
 G_DECLARE_DERIVABLE_TYPE (MetaCursorSprite,
@@ -57,6 +58,9 @@ void meta_cursor_sprite_set_texture (MetaCursorSprite *sprite,
 void meta_cursor_sprite_set_texture_scale (MetaCursorSprite *sprite,
                                            float             scale);
 
+void meta_cursor_sprite_set_texture_transform (MetaCursorSprite     *sprite,
+                                               MetaMonitorTransform  transform);
+
 CoglTexture *meta_cursor_sprite_get_cogl_texture (MetaCursorSprite *sprite);
 
 void meta_cursor_sprite_get_hotspot (MetaCursorSprite *sprite,
@@ -65,6 +69,8 @@ void meta_cursor_sprite_get_hotspot (MetaCursorSprite *sprite,
 
 float meta_cursor_sprite_get_texture_scale (MetaCursorSprite *sprite);
 
+MetaMonitorTransform meta_cursor_sprite_get_texture_transform (MetaCursorSprite *sprite);
+
 gboolean meta_cursor_sprite_is_animated (MetaCursorSprite *sprite);
 
 void meta_cursor_sprite_tick_frame (MetaCursorSprite *sprite);
