@@ -201,25 +201,23 @@ meta_kms_predict_states_in_impl (MetaKms       *kms,
                   update);
 }
 
-static gboolean
-meta_kms_update_process_in_impl (MetaKmsImpl  *impl,
+static gpointer
+meta_kms_process_update_in_impl (MetaKmsImpl  *impl,
                                  gpointer      user_data,
                                  GError      **error)
 {
   g_autoptr (MetaKmsUpdate) update = user_data;
-  gboolean ret;
-
-  ret = meta_kms_impl_process_update (impl, update, error);
+  MetaKmsFeedback *feedback;
 
+  feedback = meta_kms_impl_process_update (impl, update);
   meta_kms_predict_states_in_impl (meta_kms_impl_get_kms (impl), update);
 
-  return ret;
+  return feedback;
 }
 
-static gboolean
-meta_kms_post_update_sync (MetaKms        *kms,
-                           MetaKmsUpdate  *update,
-                           GError        **error)
+static MetaKmsFeedback *
+meta_kms_post_update_sync (MetaKms       *kms,
+                           MetaKmsUpdate *update)
 {
   meta_kms_update_seal (update);
 
@@ -227,27 +225,25 @@ meta_kms_post_update_sync (MetaKms        *kms,
                            "KMS (post update)");
 
   return meta_kms_run_impl_task_sync (kms,
-                                      meta_kms_update_process_in_impl,
+                                      meta_kms_process_update_in_impl,
                                       update,
-                                      error);
+                                      NULL);
 }
 
-gboolean
-meta_kms_post_pending_update_sync (MetaKms  *kms,
-                                   GError  **error)
+MetaKmsFeedback *
+meta_kms_post_pending_update_sync (MetaKms *kms)
 {
   return meta_kms_post_update_sync (kms,
-                                    g_steal_pointer (&kms->pending_update),
-                                    error);
+                                    g_steal_pointer (&kms->pending_update));
 }
 
-static gboolean
+static gpointer
 meta_kms_discard_pending_page_flips_in_impl (MetaKmsImpl  *impl,
                                              gpointer      user_data,
                                              GError      **error)
 {
   meta_kms_impl_discard_pending_page_flips (impl);
-  return TRUE;
+  return GINT_TO_POINTER (TRUE);
 }
 
 void
@@ -273,6 +269,8 @@ flush_callbacks (MetaKms *kms)
   GList *l;
   int callback_count = 0;
 
+  meta_assert_not_in_kms_impl (kms);
+
   for (l = kms->pending_callbacks; l; l = l->next)
     {
       MetaKmsCallbackData *callback_data = l->data;
@@ -330,13 +328,13 @@ meta_kms_flush_callbacks (MetaKms *kms)
   return callback_count;
 }
 
-gboolean
+gpointer
 meta_kms_run_impl_task_sync (MetaKms              *kms,
                              MetaKmsImplTaskFunc   func,
                              gpointer              user_data,
                              GError              **error)
 {
-  gboolean ret;
+  gpointer ret;
 
   kms->in_impl_task = TRUE;
   kms->waiting_for_impl_task = TRUE;
@@ -405,7 +403,7 @@ meta_kms_fd_impl_source_dispatch (GSource     *source,
 {
   MetaKmsFdImplSource *fd_impl_source = (MetaKmsFdImplSource *) source;
   MetaKms *kms = fd_impl_source->kms;
-  gboolean ret;
+  gpointer ret;
   GError *error = NULL;
 
   kms->in_impl_task = TRUE;
@@ -414,7 +412,7 @@ meta_kms_fd_impl_source_dispatch (GSource     *source,
                                   &error);
   kms->in_impl_task = FALSE;
 
-  if (!ret)
+  if (!GPOINTER_TO_INT (ret))
     {
       g_warning ("Failed to dispatch fd source: %s", error->message);
       g_error_free (error);
@@ -478,7 +476,7 @@ meta_kms_update_states_in_impl (MetaKms *kms)
                   NULL);
 }
 
-static gboolean
+static gpointer
 update_states_in_impl (MetaKmsImpl  *impl,
                        gpointer      user_data,
                        GError      **error)
@@ -487,14 +485,17 @@ update_states_in_impl (MetaKmsImpl  *impl,
 
   meta_kms_update_states_in_impl (kms);
 
-  return TRUE;
+  return GINT_TO_POINTER (TRUE);
 }
 
 static gboolean
 meta_kms_update_states_sync (MetaKms  *kms,
                              GError  **error)
 {
-  return meta_kms_run_impl_task_sync (kms, update_states_in_impl, NULL, error);
+  gpointer ret;
+
+  ret = meta_kms_run_impl_task_sync (kms, update_states_in_impl, NULL, error);
+  return GPOINTER_TO_INT (ret);
 }
 
 static void
