@@ -40,6 +40,7 @@
 #include "meta/meta-x11-errors.h"
 #include "meta/window.h"
 #include "x11/meta-x11-display-private.h"
+#include "x11/window-x11.h"
 
 enum
 {
@@ -97,6 +98,9 @@ struct _MetaWindowActorX11
   gboolean needs_reshape;
   gboolean recompute_focused_shadow;
   gboolean recompute_unfocused_shadow;
+  gboolean is_frozen;
+
+  cairo_rectangle_int_t client_area;
 };
 
 static MetaCullableInterface *cullable_parent_iface;
@@ -837,7 +841,6 @@ scan_visible_region (guchar         *mask_data,
 
 static void
 build_and_scan_frame_mask (MetaWindowActorX11    *actor_x11,
-                           cairo_rectangle_int_t *client_area,
                            cairo_region_t        *shape_region)
 {
   ClutterBackend *backend = clutter_get_default_backend ();
@@ -887,15 +890,28 @@ build_and_scan_frame_mask (MetaWindowActorX11    *actor_x11,
     {
       cairo_region_t *frame_paint_region, *scanned_region;
       cairo_rectangle_int_t rect = { 0, 0, tex_width, tex_height };
+      cairo_rectangle_int_t frame_rect;
+
+      /* If we update the shape regardless of the frozen state of the actor,
+       * as with Xwayland to avoid the black shadow effect, we ought to base
+       * the frame size on the buffer size rather than the reported window's
+       * frame size, as the buffer may not have been committed yet at this
+       * point.
+       */
+      if (meta_window_x11_always_update_shape (window))
+        meta_window_x11_buffer_rect_to_frame_rect (window, &rect, &frame_rect);
+      else
+        meta_window_get_frame_rect (window, &frame_rect);
 
       /* Make sure we don't paint the frame over the client window. */
       frame_paint_region = cairo_region_create_rectangle (&rect);
-      cairo_region_subtract_rectangle (frame_paint_region, client_area);
+      cairo_region_subtract_rectangle (frame_paint_region,
+                                       &actor_x11->client_area);
 
       gdk_cairo_region (cr, frame_paint_region);
       cairo_clip (cr);
 
-      meta_frame_get_mask (window->frame, cr);
+      meta_frame_get_mask (window->frame, &frame_rect, cr);
 
       cairo_surface_flush (image);
       scanned_region = scan_visible_region (mask_data, stride, frame_paint_region);
@@ -948,14 +964,13 @@ update_shape_region (MetaWindowActorX11 *actor_x11)
   MetaWindow *window =
     meta_window_actor_get_meta_window (META_WINDOW_ACTOR (actor_x11));
   cairo_region_t *region = NULL;
-  cairo_rectangle_int_t client_area;
-
-  meta_window_get_client_area_rect (window, &client_area);
 
   if (window->frame && window->shape_region)
     {
       region = cairo_region_copy (window->shape_region);
-      cairo_region_translate (region, client_area.x, client_area.y);
+      cairo_region_translate (region,
+                              actor_x11->client_area.x,
+                              actor_x11->client_area.y);
     }
   else if (window->shape_region != NULL)
     {
@@ -966,11 +981,11 @@ update_shape_region (MetaWindowActorX11 *actor_x11)
       /* If we don't have a shape on the server, that means that
        * we have an implicit shape of one rectangle covering the
        * entire window. */
-      region = cairo_region_create_rectangle (&client_area);
+      region = cairo_region_create_rectangle (&actor_x11->client_area);
     }
 
   if (window->shape_region || window->frame)
-    build_and_scan_frame_mask (actor_x11, &client_area, region);
+    build_and_scan_frame_mask (actor_x11, region);
 
   g_clear_pointer (&actor_x11->shape_region, cairo_region_destroy);
   actor_x11->shape_region = region;
@@ -1044,10 +1059,6 @@ update_opaque_region (MetaWindowActorX11 *actor_x11)
   is_maybe_transparent = is_actor_maybe_transparent (actor_x11);
   if (is_maybe_transparent && window->opaque_region)
     {
-      cairo_rectangle_int_t client_area;
-
-      meta_window_get_client_area_rect (window, &client_area);
-
       /* The opaque region is defined to be a part of the
        * window which ARGB32 will always paint with opaque
        * pixels. For these regions, we want to avoid painting
@@ -1059,7 +1070,9 @@ update_opaque_region (MetaWindowActorX11 *actor_x11)
        * case, graphical glitches will occur.
        */
       opaque_region = cairo_region_copy (window->opaque_region);
-      cairo_region_translate (opaque_region, client_area.x, client_area.y);
+      cairo_region_translate (opaque_region,
+                              actor_x11->client_area.x,
+                              actor_x11->client_area.y);
       cairo_region_intersect (opaque_region, actor_x11->shape_region);
     }
   else if (is_maybe_transparent)
@@ -1108,23 +1121,37 @@ handle_updates (MetaWindowActorX11 *actor_x11)
 {
   MetaSurfaceActor *surface =
     meta_window_actor_get_surface (META_WINDOW_ACTOR (actor_x11));
-
-  if (meta_window_actor_is_frozen (META_WINDOW_ACTOR (actor_x11)))
-    {
-      /* The window is frozen due to a pending animation: we'll wait until
-       * the animation finishes to reshape and repair the window */
-      return;
-    }
+  MetaWindow *window;
 
   if (META_IS_SURFACE_ACTOR_X11 (surface) &&
       meta_surface_actor_x11_is_unredirected (META_SURFACE_ACTOR_X11 (surface)))
     return;
 
+  window = meta_window_actor_get_meta_window (META_WINDOW_ACTOR (actor_x11));
+  if (meta_window_actor_is_frozen (META_WINDOW_ACTOR (actor_x11)))
+    {
+      /* The window is frozen due to a pending animation: we'll wait until
+       * the animation finishes to repair the window.
+       *
+       * However, with Xwayland, we still might need to update the shape
+       * region as the wl_buffer will be set to plain black on resize,
+       * which causes the shadows to look bad.
+       */
+      if (surface && meta_window_x11_always_update_shape (window))
+        {
+          update_opaque_region (actor_x11);
+          update_shape_region (actor_x11);
+        }
+
+      return;
+    }
+
   meta_surface_actor_pre_paint (surface);
 
   if (!meta_surface_actor_is_visible (surface))
     return;
 
+  meta_window_get_client_area_rect (window, &actor_x11->client_area);
   check_needs_reshape (actor_x11);
   check_needs_shadow (actor_x11);
 }
@@ -1213,6 +1240,7 @@ static void
 meta_window_actor_x11_post_paint (MetaWindowActor *actor)
 {
   MetaWindowActorX11 *actor_x11 = META_WINDOW_ACTOR_X11 (actor);
+  MetaWindow *window;
 
   actor_x11->repaint_scheduled = FALSE;
 
@@ -1238,6 +1266,14 @@ meta_window_actor_x11_post_paint (MetaWindowActor *actor)
 
       actor_x11->needs_frame_drawn = FALSE;
     }
+
+  /* This is for Xwayland, and a no-op on plain Xorg */
+  window = meta_window_actor_get_meta_window (actor);
+  if (meta_window_x11_should_thaw_after_paint (window))
+    {
+      meta_window_x11_thaw_commits (window);
+      meta_window_x11_set_thaw_after_paint (window, FALSE);
+    }
 }
 
 static gboolean
@@ -1302,6 +1338,24 @@ meta_window_actor_x11_queue_destroy (MetaWindowActor *actor)
     remove_frame_messages_timer (actor_x11);
 }
 
+static void
+meta_window_actor_x11_set_frozen (MetaWindowActor *actor,
+                                  gboolean         frozen)
+{
+  MetaWindowActorX11 *actor_x11 = META_WINDOW_ACTOR_X11 (actor);
+  MetaWindow *window = meta_window_actor_get_meta_window (actor);
+
+  if (actor_x11->is_frozen == frozen)
+    return;
+
+  actor_x11->is_frozen = frozen;
+
+  if (frozen)
+    meta_window_x11_freeze_commits (window);
+  else
+    meta_window_x11_thaw_commits (window);
+}
+
 static void
 meta_window_actor_x11_set_property (GObject      *object,
                                     guint         prop_id,
@@ -1393,9 +1447,11 @@ meta_window_actor_x11_cull_out (MetaCullable   *cullable,
                                 cairo_region_t *unobscured_region,
                                 cairo_region_t *clip_region)
 {
+  MetaWindowActorX11 *self = META_WINDOW_ACTOR_X11 (cullable);
+
   cullable_parent_iface->cull_out (cullable, unobscured_region, clip_region);
 
-  set_clip_region_beneath (META_WINDOW_ACTOR_X11 (cullable), clip_region);
+  set_clip_region_beneath (self, clip_region);
 }
 
 static void
@@ -1471,6 +1527,7 @@ meta_window_actor_x11_class_init (MetaWindowActorX11Class *klass)
   window_actor_class->pre_paint = meta_window_actor_x11_pre_paint;
   window_actor_class->post_paint = meta_window_actor_x11_post_paint;
   window_actor_class->queue_destroy = meta_window_actor_x11_queue_destroy;
+  window_actor_class->set_frozen = meta_window_actor_x11_set_frozen;
 
   actor_class->paint = meta_window_actor_x11_paint;
   actor_class->get_paint_volume = meta_window_actor_x11_get_paint_volume;
