@@ -189,7 +189,7 @@ zxdg_toplevel_v6_set_parent (struct wl_client   *client,
   MetaWindow *transient_for = NULL;
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -198,7 +198,7 @@ zxdg_toplevel_v6_set_parent (struct wl_client   *client,
       MetaWaylandSurface *parent_surface =
         surface_from_xdg_surface_resource (parent_resource);
 
-      transient_for = parent_surface->window;
+      transient_for = meta_wayland_surface_get_window (parent_surface);
     }
 
   meta_window_set_transient_for (window, transient_for);
@@ -212,7 +212,7 @@ zxdg_toplevel_v6_set_title (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -230,7 +230,7 @@ zxdg_toplevel_v6_set_app_id (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -253,7 +253,7 @@ zxdg_toplevel_v6_show_window_menu (struct wl_client   *client,
   MetaWindow *window;
   int monitor_scale;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -277,7 +277,7 @@ zxdg_toplevel_v6_move (struct wl_client   *client,
   MetaWindow *window;
   gfloat x, y;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -323,7 +323,7 @@ zxdg_toplevel_v6_resize (struct wl_client   *client,
   gfloat x, y;
   MetaGrabOp grab_op;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -344,7 +344,7 @@ zxdg_toplevel_v6_set_max_size (struct wl_client   *client,
   MetaWaylandSurfaceState *pending;
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -373,7 +373,7 @@ zxdg_toplevel_v6_set_min_size (struct wl_client   *client,
   MetaWaylandSurfaceState *pending;
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -399,7 +399,7 @@ zxdg_toplevel_v6_set_maximized (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -417,7 +417,7 @@ zxdg_toplevel_v6_unset_maximized (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -432,7 +432,7 @@ zxdg_toplevel_v6_set_fullscreen (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -453,7 +453,7 @@ zxdg_toplevel_v6_unset_fullscreen (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -467,7 +467,7 @@ zxdg_toplevel_v6_set_minimized (struct wl_client   *client,
   MetaWaylandSurface *surface = surface_from_xdg_toplevel_resource (resource);
   MetaWindow *window;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     return;
 
@@ -600,10 +600,13 @@ meta_wayland_zxdg_toplevel_v6_send_configure (MetaWaylandZxdgToplevelV6      *xd
     META_WAYLAND_SURFACE_ROLE (xdg_toplevel);
   MetaWaylandSurface *surface =
     meta_wayland_surface_role_get_surface (surface_role);
+  MetaWindow *window;
   struct wl_array states;
 
+  window = meta_wayland_surface_get_window (surface);
+
   wl_array_init (&states);
-  fill_states (&states, surface->window);
+  fill_states (&states, window);
 
   zxdg_toplevel_v6_send_configure (xdg_toplevel->resource,
                                    configuration->width,
@@ -662,7 +665,7 @@ meta_wayland_zxdg_toplevel_v6_apply_state (MetaWaylandSurfaceRole  *surface_role
   MetaRectangle old_geometry;
   gboolean geometry_changed;
 
-  window = surface->window;
+  window = meta_wayland_surface_get_window (surface);
   if (!window)
     {
       meta_wayland_surface_cache_pending_frame_callbacks (surface, pending);
@@ -845,9 +848,11 @@ static void
 scale_placement_rule (MetaPlacementRule  *placement_rule,
                       MetaWaylandSurface *surface)
 {
+  MetaWindow *window;
   int geometry_scale;
 
-  geometry_scale = meta_window_wayland_get_geometry_scale (surface->window);
+  window = meta_wayland_surface_get_window (surface);
+  geometry_scale = meta_window_wayland_get_geometry_scale (window);
 
   placement_rule->anchor_rect.x *= geometry_scale;
   placement_rule->anchor_rect.y *= geometry_scale;
@@ -885,7 +890,7 @@ finish_popup_setup (MetaWaylandZxdgPopupV6 *xdg_popup)
   xdg_popup->setup.parent_surface = NULL;
   xdg_popup->setup.grab_seat = NULL;
 
-  if (!parent_surface->window)
+  if (!meta_wayland_surface_get_window (parent_surface))
     {
       zxdg_popup_v6_send_popup_done (xdg_popup->resource);
       return;
@@ -964,6 +969,7 @@ meta_wayland_zxdg_popup_v6_apply_state (MetaWaylandSurfaceRole  *surface_role,
   MetaWaylandSurfaceRoleClass *surface_role_class;
   MetaWaylandSurface *surface =
     meta_wayland_surface_role_get_surface (surface_role);
+  MetaWindow *window;
 
   if (xdg_popup->setup.parent_surface)
     finish_popup_setup (xdg_popup);
@@ -973,7 +979,8 @@ meta_wayland_zxdg_popup_v6_apply_state (MetaWaylandSurfaceRole  *surface_role,
   surface_role_class->apply_state (surface_role, pending);
 
   /* If the window disappeared the surface is not coming back. */
-  if (!surface->window)
+  window = meta_wayland_surface_get_window (surface);
+  if (!window)
     return;
 
   if (!pending->newly_attached)
@@ -988,7 +995,7 @@ meta_wayland_zxdg_popup_v6_apply_state (MetaWaylandSurfaceRole  *surface_role,
 
       window_geometry =
         meta_wayland_zxdg_surface_v6_get_window_geometry (xdg_surface);
-      meta_window_wayland_finish_move_resize (surface->window,
+      meta_window_wayland_finish_move_resize (window,
                                               window_geometry,
                                               pending);
     }
@@ -1013,7 +1020,8 @@ meta_wayland_zxdg_popup_v6_configure (MetaWaylandShellSurface        *shell_surf
     META_WAYLAND_ZXDG_POPUP_V6 (shell_surface);
   MetaWaylandZxdgSurfaceV6 *xdg_surface =
     META_WAYLAND_ZXDG_SURFACE_V6 (xdg_popup);
-  MetaWindow *parent_window = xdg_popup->parent_surface->window;
+  MetaWindow *parent_window =
+    meta_wayland_surface_get_window (xdg_popup->parent_surface);
   int geometry_scale;
   int x, y;
 
@@ -1029,8 +1037,8 @@ meta_wayland_zxdg_popup_v6_configure (MetaWaylandShellSurface        *shell_surf
     return;
 
   geometry_scale = meta_window_wayland_get_geometry_scale (parent_window);
-  x = (configuration->x - parent_window->rect.x) / geometry_scale;
-  y = (configuration->y - parent_window->rect.y) / geometry_scale;
+  x = configuration->rel_x / geometry_scale;
+  y = configuration->rel_y / geometry_scale;
 
   zxdg_popup_v6_send_configure (xdg_popup->resource,
                                 x, y,
@@ -1048,7 +1056,8 @@ meta_wayland_zxdg_popup_v6_managed (MetaWaylandShellSurface *shell_surface,
 
   g_assert (parent);
 
-  meta_window_set_transient_for (window, parent->window);
+  meta_window_set_transient_for (window,
+                                 meta_wayland_surface_get_window (parent));
   meta_window_set_type (window, META_WINDOW_DROPDOWN_MENU);
 }
 
@@ -1328,7 +1337,7 @@ meta_wayland_zxdg_surface_v6_apply_state (MetaWaylandSurfaceRole  *surface_role,
     meta_wayland_zxdg_surface_v6_get_instance_private (xdg_surface);
   MetaWaylandSurface *surface =
     meta_wayland_surface_role_get_surface (surface_role);
-  MetaWindow *window = surface->window;
+  MetaWindow *window = meta_wayland_surface_get_window (surface);
   MetaWaylandSurfaceRoleClass *surface_role_class;
 
   surface_role_class =
